//
// FEB98.CPP
//
// This program demonstrates a problem with Visual C++
// 5.0.  One of the features of the Standard Template
// Library is a special iterator type, the output iterator.
// This program is supposed to show how one might use
// an output iterator to display a desired attribute
// of an object.
//
// The output iterators are designed to print the value
// of the m_Bar member of a Foo object.  The goal is
// to do this regardless of whether a Foo object is
// passed, or a pointer to a Foo object.
//
// Unfortunately, VC++ 5.0 seems to want to print the
// raw value of the pointer instead of using the
// ostream insertion function defined for foo.  So while
// this program is supposed to print four lines in a row
// with the same value, VC++ on my machine gives this
// output:
//
//  42
//  42
//  42
//  0064FDE4
//
// Note that Borland C++ and GNU g++ both handle this
// task properly.
//
#include <iostream>
#include <iterator>

using namespace std;

class Foo
{
public :
    Foo( int i )
    {
        m_Bar = i;
    }
    int m_Bar;
};

ostream& operator<<( ostream& s, const Foo *f )
{
    return s << f->m_Bar;
}

ostream& operator<<( ostream& s, const Foo &f )
{
    return s << f.m_Bar;
}

main()
{
    Foo f( 42 );
    cout << f << "\n";
    cout << &f << "\n";
    ostream_iterator<Foo> o1( cout, "\n" );
    *o1 = f;
    o1++;
    ostream_iterator<Foo *> o2( cout, "\n" );
    *o2 = &f;
    o2++;
    return 1;
}
