//------------------------------------------------------------------
//  Ports.c - TCP port scanner to list active ports. Ton Plooy 1997
//  Compile and link with 'cl ports.c -link wsock32.lib icmp.lib'
//------------------------------------------------------------------
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#define  WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <winsock.h>
#include "icmp.h"

#define LINELEN                     16
#define MAX_TIMEOUT                 1500

//------------------------------------------------------------------
int   CheckPort(DWORD dwIP, USHORT usPort, ULONG timeout);
int   connectex(SOCKET s, const struct sockaddr *name,
                int namelen, long timeout);
int   closesocketex(SOCKET sock, BOOL reset);
DWORD stoip(char *lpIP);
ULONG ping(DWORD dwIP, ULONG ulTimeOut);
void  HexDump(char *Data, int nLen);

//------------------------------------------------------------------
//
//------------------------------------------------------------------

main(int argc, char *argv[])
{
    int          i, tries = 0;
    BOOL         bStop = FALSE;
    DWORD        dwIP, dwRead, p, missed = 0, time = 0;
    WSADATA      wsaData;
    HANDLE       hCon;
    INPUT_RECORD Rec;

    if (argc < 2) {
        printf("Usage: ports <ip-address>\n");
        return 0;
    }
    WSAStartup(0x0101, &wsaData);

    // Resolve machine name or translate address
    if ((dwIP = stoip(argv[1])) == 0) {
        printf("Error: address %s unknown\n", argv[1]);
        WSACleanup();
        return 0;
    }
    hCon = GetStdHandle(STD_INPUT_HANDLE);

    // Calulate avarage echo reply time over four pings
    do {

        if ((p = ping(dwIP, MAX_TIMEOUT)) == 0)
            missed++;
        else
            time += p;

    } while (++tries < 4);

    if (time == 0) {
        printf("No reply from %s...\n", argv[1]);
    }
    else {
        // At least one response seen, continue
        time = (time + (missed * MAX_TIMEOUT)) / tries;
        printf("Avarage reply time from %s is %lu msecs\n",
                argv[1], time);

        // Set port connect timeout at ping time plus 50 %
        time = min(time + (time / 2), MAX_TIMEOUT);

        printf("Starting portscan (press space to stop)...\n");

        for (i = 1; (i <= 1024) && (bStop == FALSE); i++) {

            if ((i % 128) == 0)
                printf("Scanning port %4d...\n", i);

            CheckPort(dwIP, (USHORT)i, time);

            // Check keyboard input, stop if space bar is pressed
            PeekConsoleInput(hCon, &Rec, 1, &dwRead);
            if ((dwRead > 0) && (Rec.EventType == KEY_EVENT)) {
                ReadConsoleInput(hCon, &Rec, 1, &dwRead);
                if (Rec.Event.KeyEvent.wVirtualKeyCode == ' ')
                    bStop = TRUE;
            }
        }
    } 

    WSACleanup();

    return 0;
}

//------------------------------------------------------------------

int CheckPort(DWORD dwIP, USHORT usPort, ULONG timeout)
{
    // Check if the IP port is listening on the given machine
    int                 nLen;
    unsigned char       Data[512];
    SOCKET              s;
    fd_set              FdSet;
    struct timeval      Time = { 1, 0 };
    struct sockaddr_in  SrvAddr, CliAddr;

    if ((s = socket(PF_INET, SOCK_STREAM, 0)) == INVALID_SOCKET) {
        printf("Error: unable to create socket\n");
        return 0;
    }

    // Bind the socket
    CliAddr.sin_family      = AF_INET;
    CliAddr.sin_addr.s_addr = INADDR_ANY;
    CliAddr.sin_port        = 0;
    bind(s, (PSOCKADDR)&CliAddr, sizeof(CliAddr));

    // Connect
    SrvAddr.sin_family      = AF_INET;
    SrvAddr.sin_addr.s_addr = dwIP;
    SrvAddr.sin_port        = htons(usPort);
    if (connectex(s, (LPSOCKADDR)&SrvAddr, sizeof(SrvAddr),
                                timeout) == SOCKET_ERROR) {
        closesocketex(s, TRUE);
        return 0;
    }
    // Connect succeeded, so the port is listening
    printf("Port %d is listening\n", usPort);

    // Send some data to the port to see if we get anything
    // back. If we do, hexdump the received data.
    memset(Data, 0, sizeof(Data));
    Data[0] = 'X'; // Send just one byte
    if (send(s, Data, 1, 0) > 0) {
        FD_ZERO(&FdSet);
        FD_SET(s, &FdSet);
        Time.tv_sec  = timeout / 1000L;
        Time.tv_usec = (timeout % 1000) * 1000;
        if ((select(0, &FdSet, NULL, NULL, &Time) > 0) &&
                ((nLen = recv(s, Data, sizeof(Data), 0)) > 0)) {
            HexDump(Data, nLen);
        }
    }
    closesocket(s);

    return 1;
}

//------------------------------------------------------------------

int connectex(SOCKET s, const struct sockaddr *name,
                            int namelen, long timeout)
{
    // As connect() but with timeout setting.
    int            rc = 0;
    ULONG          ulB;
    struct timeval Time;
    fd_set         FdSet;

    ulB = TRUE; // Set socket to non-blocking mode
    ioctlsocket(s, FIONBIO, &ulB);

    if (connect(s, name, sizeof(SOCKADDR)) == SOCKET_ERROR) {

        if (WSAGetLastError() == WSAEWOULDBLOCK) {
            // now wait for the specified time
            FD_ZERO(&FdSet);
            FD_SET(s, &FdSet);

            Time.tv_sec  = timeout / 1000L;
            Time.tv_usec = (timeout % 1000) * 1000;
            rc = select(0, NULL, &FdSet, NULL, &Time);

        }
    }

    ulB = FALSE; // Restore socket to blocking mode
    ioctlsocket(s, FIONBIO, &ulB);

    return (rc > 0) ? 0 : SOCKET_ERROR;
}

//------------------------------------------------------------------

int closesocketex(SOCKET sock, BOOL reset)
{
    // Close socket. If reset is true we do a hard close.
    int    rc;
    LINGER linger;

    if (sock == INVALID_SOCKET)
        return 0;

    if (reset == TRUE) {
        // Enable linger with a zero timeout
        linger.l_onoff  = TRUE;
        linger.l_linger = 0;
        setsockopt(sock, SOL_SOCKET, SO_LINGER,
                    (char FAR *)&linger, sizeof(linger));
    }
    rc = closesocket(sock);

    return rc;
}

//------------------------------------------------------------------

DWORD stoip(char *lpIP)
{
    // String to IP: optionally resolve a host name and convert
    // an ASCII IP dot address to its binary network form.
    int   i, bDot = TRUE;
    char  szBuffer[64];
    DWORD dwIP = 0L;

    if ((lpIP[0] != '\0') && (isdigit(lpIP[0]) == FALSE)) {
        // Name is specified, lookup ip address
        struct hostent *pHost;

        if ((pHost = gethostbyname((LPSTR)lpIP)) != NULL) {
            memcpy((char *)&dwIP, pHost->h_addr_list[0], 4);
        }
    }
    else {
        // Address specified. Remove all preceding zeros first,
        // inet_adr will interpret it as an octal number.
        i = 0;
        while (*lpIP) {

            while ((*(lpIP) == '0') && (isdigit(*(lpIP + 1)))
                                            && (bDot == TRUE)) {
                lpIP++;
            }
            bDot = (*lpIP == '.');
            szBuffer[i++] = *lpIP++;
        }
        szBuffer[i] = '\0';
        dwIP = inet_addr(szBuffer);
    }
    return dwIP;
}

//------------------------------------------------------------------

ULONG ping(DWORD dwIP, ULONG ulTimeOut)
{
    ULONG                 ulTime = 0;
    HANDLE                hIcmp;
    struct _PongData {
        PONGHDR           PongHdr;
        char              PongData[20];
    } Pong;
    static char           PingData[20] = "Crunch Technologies";

    hIcmp = IcmpCreateFile();
    if (IcmpSendEcho(hIcmp, dwIP,
                    PingData, (WORD)(strlen(PingData) + 1), NULL,
                    (char *)&Pong, sizeof(Pong), ulTimeOut) == 1) {
        // Return at least 1 msec
        ulTime = max(Pong.PongHdr.dwTime, 1);
    }
    IcmpCloseHandle(hIcmp);

    return ulTime;
}

//------------------------------------------------------------------

void HexDump(char *Data, int nLen)
{
    // Do a hexdump with the supplied data to stdout
    int           i, nCount;
    unsigned char Line[LINELEN];
    ULONG         ulOffset = 0L;

    while (nLen > 0L) {
        nCount = nLen > LINELEN ? LINELEN : nLen;
        memcpy(Line, Data + ulOffset, nCount);
        printf("%04lX  ",  ulOffset);
        for (i = 0; i < LINELEN; i++) {
            (i < nCount) ? printf("%02X ", (unsigned int)Line[i]) :
                            printf("   ");
        }
        printf(" ");
        for (i = 0; (i < LINELEN) && (i < nCount); i++) {
            printf("%c", isprint(Line[i]) ? Line[i] : '.');
        }
        printf("\n");
        nLen     -= nCount;
        ulOffset += nCount;
    }
    return ;
}

