// wdj.c

#include "ntddk.h"
#include "wdj.h"

int sprintf(char *s, const char *format, ...);

typedef struct _DEVICE_EXTENSION {
    ULONG Ordinal;
} DEVICE_EXTENSION, *PDEVICE_EXTENSION;

NTSTATUS WdjDispatch(IN PDEVICE_OBJECT, IN PIRP);
VOID WdjUnload(IN PDRIVER_OBJECT);

//-----------------------------------------------------------------
NTSTATUS DriverEntry(IN PDRIVER_OBJECT  DriverObject,
                     IN PUNICODE_STRING RegistryPath)
{
    UNICODE_STRING symName, deviceName;
    PDEVICE_OBJECT deviceObject0 = NULL, deviceObject1 = NULL;
    PDEVICE_EXTENSION deviceExt;
    NTSTATUS status;

    UNREFERENCED_PARAMETER(RegistryPath);

    KdPrint(("WDJ!DriverEntry: Driver loading\n"));

    //
    // Create a normal device object with a normal symbolic link
    //

    RtlInitUnicodeString(&deviceName, L"\\Device\\Wdj0");
    RtlInitUnicodeString(&symName, L"\\??\\UserWdj0");
                     
    status = IoCreateDevice(DriverObject, sizeof(DEVICE_EXTENSION),
                            &deviceName, FILE_DEVICE_WDJ, 0,
                            FALSE, &deviceObject0);

    if (NT_SUCCESS(status)) {
        deviceExt = (PDEVICE_EXTENSION)deviceObject0
                ->DeviceExtension;
        deviceExt->Ordinal = 0;
        status = IoCreateSymbolicLink(&symName, &deviceName);
        if (!NT_SUCCESS(status)) {
            IoDeleteDevice(deviceObject0);
            deviceExt->Ordinal = -1;  // no symbolic link created
        }
    }

    //
    // Create an exclusive device object without a symbolic link
    //

    RtlInitUnicodeString(&deviceName, L"\\Device\\Wdj1");
    RtlInitUnicodeString(&symName, L"\\??\\UserWdj1");
                     
    status = IoCreateDevice(DriverObject, sizeof(DEVICE_EXTENSION),
                            &deviceName, FILE_DEVICE_WDJ, 0,
                            TRUE, &deviceObject1);

    if (NT_SUCCESS(status)) {   
        deviceExt = (PDEVICE_EXTENSION)deviceObject1
                ->DeviceExtension;
        deviceExt->Ordinal = -1;  // no symbolic link created
    }

    //
    // Fill in the dispatch table.
    //

    DriverObject->DriverUnload                 = WdjUnload;
    DriverObject->MajorFunction[IRP_MJ_CREATE] = WdjDispatch;
    DriverObject->MajorFunction[IRP_MJ_CLOSE]  = WdjDispatch;

    return STATUS_SUCCESS;

} // DriverEntry

//-----------------------------------------------------------------
NTSTATUS WdjDispatch(IN PDEVICE_OBJECT DeviceObject,
                     IN PIRP           Irp)
{
    NTSTATUS status = STATUS_SUCCESS;
    PIO_STACK_LOCATION irpStack = IoGetCurrentIrpStackLocation(Irp);

    KdPrint(("WDJ!WdjDispatch\n"));

    switch (irpStack->MajorFunction) {
        case IRP_MJ_CREATE:
            KdPrint(("  WDJ: IRP_MJ_CREATE\n"));
            if (irpStack->FileObject->FileName.Length != 0) {
                status = STATUS_INVALID_PARAMETER;
            }
            break;

        case IRP_MJ_CLOSE:
            KdPrint(("  WDJ: IRP_MJ_CLOSE\n"));
            break;
    }

    Irp->IoStatus.Status = status;
    Irp->IoStatus.Information = 0;
    IoCompleteRequest(Irp, IO_NO_INCREMENT);
    return status;

} // WdjDispatch


//-----------------------------------------------------------------
VOID WdjUnload(IN PDRIVER_OBJECT DriverObject)
{
    CHAR ansiBuffer[64];
    UNICODE_STRING uniName;
    ANSI_STRING ansiName;
    PDEVICE_OBJECT deviceObject, nextDevice;
    PDEVICE_EXTENSION deviceExt;

    KdPrint(("WDJ!WdjUnload: Driver unloading\n"));
    
    //
    // Traverse device object list, free any resources.
    //

    deviceObject = DriverObject->DeviceObject;
    
    while (deviceObject) {
        nextDevice = deviceObject->NextDevice;
        deviceExt = (PDEVICE_EXTENSION)deviceObject->DeviceExtension;
        if (deviceExt->Ordinal != -1) {
            sprintf(ansiBuffer, "\\??\\UserWdj%d", 
                    deviceExt->Ordinal);
            RtlInitAnsiString(&ansiName, ansiBuffer);
            RtlAnsiStringToUnicodeString(&uniName, &ansiName, TRUE); 
            IoDeleteSymbolicLink(&uniName);
            RtlFreeUnicodeString(&uniName);
        }
        IoDeleteDevice(deviceObject);
        deviceObject = nextDevice;
    }

} // WdjUnload
